.str{color:#718c00}.kwd{color:#8959a8}.com{color:#8e908c}.typ,.dec{color:#f5871f}.lit,.atn{color:#eab700}.pun,.opn,.clo,.pln{color:#4d4d4c}.tag,.var,.fun{color:#c82829}.atv{color:#3e999f}pre.prettyprint{border:1px solid #d1d1d1;padding:1em;padding-bottom:1.5em;padding-top:1.5em}
